from random import random, choice, seed

from py_school_match.entities.characteristic import Characteristic


def simulate_characteristics(students, schools, criteria_prob):
    for student in students:
        # This is VERY important, as the criteria_prob is a dict, any difference in this order is will be reflected in other random values
        sorted_criteria_prob = sorted(criteria_prob, key=lambda c: criteria_prob[c])

        for criteria in sorted_criteria_prob:
            if random() < criteria_prob[criteria]:
                value = True if criteria.criteria_type == bool else [choice(schools)]

                characteristic = Characteristic(criteria, value)
                student.add_characteristic(characteristic)


"""This module implements the *'Deferred Acceptance'* algorithm.
This algorithm comes from Gale and Shapley (1962), and the
implementation is based on Erdil and Ergin (2006)."""

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm


class DA(AbstractMatchingAlgorithm):
    """This class implements the *Deferred Acceptance* algorithm.
    
    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate priorities).
    This works by 'proposing' students to their most preferred
    (and available) school. Schools, on the other hand, accept
    or reject these students based on their priority.
    """

    def run(self, students, schools, ruleset):
        """Runs the *Deferred Acceptance* algorithm.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """
        for student in students:
            setattr(student, 'option_n', 0)
            setattr(student, 'assigned', False)

        still_free_students = True
        while still_free_students:

            still_free_students = False
            for student in students:

                if not student.assigned:
                    pref_school = student.preferences[student.option_n]
                    DA.assign_student(student, pref_school)

                    rejected_student = pref_school.assignation.get_rejected_student()

                    if rejected_student:
                        rejected_student.option_n += 1

                        if rejected_student.option_n < len(rejected_student.preferences):
                            DA.unassign_student(rejected_student)
                            still_free_students = True
                        else:
                            DA.definitely_unassign_student(rejected_student)

    @staticmethod
    def assign_student(student, school):
        """Assigns a student to a school."""
        student.assigned_school = school
        student.assigned = True
        school.assignation.append(student)

    @staticmethod
    def unassign_student(student):
        """Unassigns a student to a school."""
        student.assigned = False
        student.assigned_school = None

    @staticmethod
    def definitely_unassign_student(student):
        """Sets a student as rejected by all schools."""
        student.assigned = True
        student.assigned_school = None



"""This module defines a ruleset."""


class RuleSet:
    """This class defines ruleset. 
    A ruleset is a set of rules which are prioritized.
    """

    def __init__(self, new_rules=None):
        """Initializes a ruleset.

        :param new_rules: A list of rules.
        :type new_rules: list.
        """
        self.rules = new_rules if new_rules else []

    def add_rule(self, rule):
        """Adds a rule from the ruleset."""
        self.rules.append(rule)

    def add_rule_in_position(self, rule, position):
        """Adds a rule from the ruleset."""
        self.rules.insert(position, rule)

    def remove_rule(self, rule):
        """Removes a rule from the ruleset."""
        self.rules.remove(rule)

    def get_rules_prioritized(self):
        """Returns the rules of the ruleset."""
        return self.rules






"""This module defines a slot."""

from itertools import groupby


class Slot:
    """This class defines a slot.
    A slot is used to store and prioritize students
    inside a school.
    """

    def __init__(self, rule, capacity):
        """Initializes a slot.

        :param rule: The rule used in the slot.
        :type rule: :class:`.Rule`..
        :param capacity: The slot capacity.
        :type capacity: int.
        """
        self.rule = rule
        self.capacity = capacity
        self.queue = []

    def is_general_slot(self):
        """If the slot has no rule attached."""
        return self.rule is None

    def is_quoted_slot(self):
        """If the slot has a quota."""
        return not self.is_general_slot()

    def append(self, student):
        """Appends a student to the slot."""
        self.queue.append(student)

    def sort_slot(self, get_priority):  # ToDo: Optimize insert + sort (the list is already sorted before the insert).
        """Sorts the slot according to the rules."""
        self.queue.sort(key=lambda student: get_priority(self, student), reverse=True)

    def gen_indifference_groups(self, get_priority):
        """Generates groups where all students have the same priority."""
        self.sort_slot(get_priority)
        groups = groupby(self.queue, key=lambda student: get_priority(self, student))
        groups_values = [[item for item in data] for (key, data) in groups]
        return groups_values

    def pop_rejected(self):
        """Removes rejected students from the slot."""
        if len(self.queue) > self.capacity:
            rejected = self.queue[-1]
            del self.queue[-1]
            return rejected
        else:
            return None




"""This module implements the algorithm *'Top Trading Cycles'*.
This implementation is based on Abdulkadiroglu and Sonmez (2003).

.. Note::
   This algorithm is considerably slower than the others. 
"""

from graph_tool.all import Graph, graph_draw, all_circuits, arf_layout

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm
from py_school_match.entities.student_queue import StudentQueue
from py_school_match.util.file_util import gen_filepath


class TTC(AbstractMatchingAlgorithm):
    """This class searches for cycles where each student gets his best option.

    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate priorities).
    This works by generating a directed graph, where each student points
    at at his best option, and each school points at the student (or students)
    with the highest priority.
    """

    EDGE_WIDTH_SIZE_FACTOR = 700
    """Size factor (in the image) of each edge that is not part of the main cycle."""
    EDGE_WIDTH_CYCLE_SIZE = 10
    """Size factor (in the image) of each edge that takes part of the main cycle."""

    def __init__(self, generate_images=False, images_folder="TTC_images", use_longest_cycle=True):
        """Initializes the algorithm.

        :param generate_images: If the process generates images or not.
        :type generate_images: bool
        :param images_folder: Where images are saved.
        :type images_folder: str
        :param use_longest_cycle: If the algorithm applies the longest cycle available, or the first one encountered.
        :type use_longest_cycle: bool
        """
        self.generate_images = generate_images
        self.images_folder = images_folder
        self.use_longest_cycle = use_longest_cycle

        self.__graph = None
        self.__vertices_by_school_id = None
        self.__vertices_by_student_id = None
        self.__students_by_id = None
        self.__schools_by_id = None

        self.__entity_id = None
        self.__entity_type = None

    def reset_variables(self):
        """Resets all variables."""
        self.__graph = Graph()
        self.__vertices_by_school_id = {}
        self.__vertices_by_student_id = {}
        self.__students_by_id = {}
        self.__schools_by_id = {}

        self.__entity_id = self.__graph.new_vertex_property("int")
        self.__graph.vertex_properties["entity_id"] = self.__entity_id

        self.__entity_type = self.__graph.new_vertex_property("string")
        self.__graph.vertex_properties["entity_type"] = self.__entity_type

    def run(self, students, schools, ruleset):
        """Runs the algorithm.
        First it creates the graph, then it lists all the cycles available,
        after that it selects one cycle, and applies it. Finally, it starts
        the process again.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """
        self.reset_variables()

        can_improve = True
        iteration_counter = 1

        while can_improve:

            self.structure_graph(students, schools)

            cycles = [c for c in all_circuits(self.__graph, unique=True)]
            # print("CYCLES", cycles, "iteration", iteration_counter)

            cycle_edges = []

            if cycles:
                for cycle in cycles: # ToDo: Possible optimisation: apply all disjoint cycles at once
                    for current_v_index in range(len(cycle)):
                        next_v_index = (current_v_index + 1) % len(cycle)

                        from_v = self.__graph.vertex(cycle[current_v_index])
                        target_v = self.__graph.vertex(cycle[next_v_index])
                        edge = self.__graph.edge(from_v, target_v)
                        cycle_edges.append(edge)

                        if self.__entity_type[from_v] == "st":
                            sel_student = self.__students_by_id[self.__entity_id[from_v]]
                            sel_school = self.__schools_by_id[self.__entity_id[target_v]]
                            sel_student.assigned_school = sel_school
                            sel_school.assignation.append(sel_student)

                        # vertex_school_target_id = self.__entity_id[target_v]
                        # vertex_school_target = self.__schools_by_id[vertex_school_target_id]

                        # print("CYCLE: Student", sel_student.id, "School", sel_school.id)

                        # print("VVV: School {} -> School {}    (Student {}) ".format(self.__entity_id[from_v], self.__entity_id[target_v], self.__entity_id[self.__graph.edge(from_v, target_v)]))

                    if self.generate_images:
                        self.generate_image(cycle_edges, iteration_n=iteration_counter)
            else:
                can_improve = False

            self.__graph.clear()
            iteration_counter += 1

    def structure_graph(self, students, schools):
        """Creates a graph where students points to schools, and schools points to students.

        In the graph, each student points at at his best option, and each school points
        at the student (or students) with the highest priority.

        :param students: List of students.
        :type students: list
        :param schools: 
        :type schools: list
        """
        if not self.__students_by_id and not self.__schools_by_id:
            for student in students:
                self.__students_by_id[student.id] = student
            for school in schools:
                self.__schools_by_id[school.id] = school

        for school in schools:
            setattr(school, 'preferences', StudentQueue(school, preference_mode=True))

        remaining_students = [student for student in students if not student.assigned_school]

        for student in remaining_students:
            for pref_school in student.preferences:
                pref_school.preferences.append(student)

        for student in remaining_students:
            v_source_student = self.create_vertex_student(student)

            pref_school = next((school for school in student.preferences if len(school.assignation.get_all_students()) < school.capacity), None)

            if pref_school:
                v_target_school = self.create_vertex_school(pref_school)
                self.create_edge(v_source_student, v_target_school)

        for school in schools:
            if len(school.assignation.get_all_students()) < school.capacity:
                v_source_school = self.create_vertex_school(school)

                pref_student = next(iter(school.preferences.get_all_students()), None)

                if pref_student:
                    v_target_student = self.create_vertex_student(pref_student)
                    self.create_edge(v_source_school, v_target_student)

        # graph_draw(self.__graph,
        #            vertex_text=self.__entity_id, vertex_shape="circle",
        #            output_size=(1000, 1000), bg_color=[1., 1., 1., 1], output="graph.png")

    def create_vertex_student(self, student):
        """Defines a new student as a vertex in the graph (if it did not existed before)."""
        if student.id in self.__vertices_by_student_id:
            vertex = self.__vertices_by_student_id[student.id]
        else:
            vertex = self.__graph.add_vertex()
            self.__vertices_by_student_id[student.id] = vertex
            self.__entity_id[vertex] = student.id
            self.__entity_type[vertex] = "st" # ToDo: There may be other ways to do this.
        return vertex

    def create_vertex_school(self, school):
        """Defines a new school as a vertex in the graph (if it did not existed before)."""
        if school.id in self.__vertices_by_school_id:
            vertex = self.__vertices_by_school_id[school.id]
        else:
            vertex = self.__graph.add_vertex()
            self.__vertices_by_school_id[school.id] = vertex
            self.__entity_id[vertex] = school.id
            self.__entity_type[vertex] = "sc"
        return vertex

    def create_edge(self, source_v, target_v):
        """Creates a directed edge between two vertices."""
        self.__graph.add_edge(source_v, target_v)

    def generate_image(self, cycle_edges, iteration_n=0):
        """Generates an image of a graph.

        :param cycle_edges: Edges which are part of the main cycle (they will be highlighted in red).
        :type cycle_edges: list
        :param iteration_n: Number of iteration of the algorithm (this is added in the filename of the image).
        :type iteration_n: int

        .. DANGER::
        This is an experimental feature.
        """
        edge_color = self.__graph.new_edge_property("vector<float>")
        edge_width = self.__graph.new_edge_property("int")

        for edge in self.__graph.edges():
            if edge in cycle_edges:
                edge_color[edge] = [1., 0.2, 0.2, 0.999]
                edge_width[edge] = 7
            else:
                edge_color[edge] = [0., 0., 0., 0.3]
                edge_width[edge] = 4

        vertex_shape = self.__graph.new_vertex_property("string")
        vertex_size = self.__graph.new_vertex_property("int")

        for vertex in self.__graph.vertices():
            if self.__entity_type[vertex] == "st":
                vertex_shape[vertex] = "circle"
                vertex_size[vertex] = 1
            else:
                vertex_shape[vertex] = "double_circle"
                vertex_size[vertex] = 100

        # pos = sfdp_layout(self.__graph, C=10, p=5, theta=2, gamma=1)
        pos = arf_layout(self.__graph, d=0.2, a=3)
        graph_draw(self.__graph, pos=pos, vertex_text=self.__entity_id, vertex_font_size=1,  # ToDo: Move image related code outside the class.
                   vertex_fill_color=[0.97, 0.97, 0.97, 1], vertex_color=[0.05, 0.05, 0.05, 0.95],
                   vertex_shape=vertex_shape, edge_color=edge_color, edge_pen_width=edge_width,
                   output_size=(1000, 1000), bg_color=[1., 1., 1., 1], output=self.generate_filename(iteration_n))

    def generate_filename(self, iteration_n):  # ToDo: Move this to utils
        """Returns a filename (which is used to generate the images)."""
        filename = "Graph (iteration {})".format(iteration_n) if iteration_n > 0 else "Graph"
        output_file = gen_filepath(self.images_folder, filename=filename, extension="png")
        return output_file



"""This module defines a criteria."""


class Criteria:
    """This class defines a criteria.
    The goal of a criteria is to associate an adjective
    and a value type.
    """

    def __init__(self, name, criteria_type):
        """Initializes a characteristic.

        :param name: The adjective's name.
        :type name: str.
        :param criteria_type: The type of the adjective
        :type criteria_type: type.
        """
        self.name = name
        self.criteria_type = criteria_type



"""This module implements methods to visualize
assignations easily
"""

from collections import defaultdict


def get_positions_stat(planner):
    """Returns a dictionary with the positions."""
    ranking_stat = defaultdict(int)
    for student in planner.students:
        if student.assigned_school:
            ranking = student.preferences.index(student.assigned_school) + 1
            ranking_stat[ranking] += 1
        else:
            ranking_stat["NA"] += 1

    return ranking_stat


def display_assignation(positions):
    """Shows the assignation."""
    for r in positions:
        for s in range(positions[r]):
            if r != "NA":
                print(r + 1)
            else:
                print("NA")


def display_assignation_from_planner(planner):
    """Shows an assignation."""
    ranking_stat = get_positions_stat(planner)
    display_assignation(ranking_stat)


def display_positions(positions):
    """Shows the positions."""
    if "NA" in positions:
        print(positions["NA"], end=", ")
    else:
        print(0, end=", ")
    positions_n = [str(positions[r]) for r in positions if r != "NA"]

    print(", ".join(positions_n))


def display_positions_from_planner(planner):
    """Shows the positions."""
    ranking_stat = get_positions_stat(planner)
    display_positions(ranking_stat)


def save_positions(positions, algorithm, alpha, iteration, file="distribution.csv"):
    """Saves positions to a file. Algorithms, alpha and iteration are used as text."""
    positions_n = []

    if "NA" in positions:
        positions_n.append(positions["NA"])
        del positions["NA"]
    else:
        positions_n[0].append(0)

    positions_n.extend([positions[r] for r in sorted(positions)])

    with open(file, "a+") as f:
        line_desc = "{}, {}, {}".format(algorithm, alpha, iteration)
        line_values = ", ".join([str(p) for p in positions_n])
        line_text = "{}, {}".format(line_desc, line_values)

        f.write(line_text + "\n")


def save_data_from_planner(planner, algorithm, alpha, iteration, file="all_distributions.csv"):
    """Saves positions from a social planner to a file. Algorithms, alpha and iteration are used as text."""
    ranking_stat = get_positions_stat(planner)
    save_positions(ranking_stat, algorithm, alpha, iteration, file)




"""This module defines an abstract class for other algorithms."""

from abc import ABCMeta, abstractmethod


class AbstractMatchingAlgorithm(metaclass=ABCMeta):
    """Abstract class for matching algorithms.
    Defined, using ABCMeta.
    """

    @abstractmethod
    def run(self, students, schools, ruleset):
        """
        Abstract method algorithms should implement.
        
        :param students: List of students.
        :type students: list
        :param schools: List of schools.
        :type schools: list
        :param ruleset: The ruleset to be used.
        :type ruleset: Ruleset
        """
        pass





"""This module defines a school."""

from itertools import count

from py_school_match.entities.student_queue import StudentQueue


class School:
    """This class defines school."""

    __id_counter = count(0)
    """This is used to generate an incremental id."""

    def __init__(self, capacity, id_=""):
        """Initializes a school.

        :param capacity: The maximum number of students.
        :type capacity: int.
        :param id_: Any unique identifier.
        :type id_: Any.
        """
        self.id = id_ if id_ else next(School.__id_counter)
        self.capacity = capacity

        self.ruleset = None
        self.assignation = StudentQueue(self)

    def reset_assignation(self):
        """Resets the assignations of the school."""
        self.assignation = StudentQueue(self)

    def set_ruleset_n_reset(self, ruleset):
        """Sets a new ruleset and resets the school's assignation."""
        self.ruleset = ruleset
        self.reset_assignation()

    def set_ruleset(self, ruleset):
        """Sets a new ruleset to the school."""
        self.ruleset = ruleset

    def get_ruleset(self):
        """Returns the school's ruleset."""
        return self.ruleset

    @staticmethod
    def reset_ids():
        """Resets the incremental id."""
        School.__id_counter = count(0)



"""This module defines a characteristic.
The goal of a characteristic is to relate a criteria
and a value."""


class Characteristic:
    """This class define a characteristic.

    Characteristics are used to define which criteria
    does a student have (and what value does it contain).
    """

    def __init__(self, criteria, value, recognized_only_in=None):
        """Initializes a characteristic.

        :param criteria: A criteria.
        :type criteria: :class:`.Criteria`.
        :param value: The value associated with the criteria.
        :type value: Any.
        :param recognized_only_in: Where the characteristic is valid.
        :type recognized_only_in: list
        """
        self.criteria = criteria
        self.value = value

        self.__recognized_in = [recognized_only_in] if recognized_only_in is not None else []

    @property
    def recognizers(self):
        return self.__recognized_in

    def is_recognized_in(self, school):
        """Checks if the characteristic is recognized in the school.
        If the characteristic does not have any recognizer specified,
        the characteristic is recognized at every school."""
        return school in self.__recognized_in or not self.__recognized_in

    def add_recognizer(self, recognizer):
        """Adds a school where the characteristic is recognized."""
        self.__recognized_in.append(recognizer)



"""This module implements the algorithm *'Non-stable improvement
cycles'*. This means, it applies *'Deferred Acceptance with single
tie breaking'*, and then searches for non-stable cycles.
"""

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm
from py_school_match.algorithms.da_stb import DASTB
from py_school_match.algorithms.pi import PI


class NSIC(AbstractMatchingAlgorithm):
    """This class implements the 'Non-stable improvement
    cycles'* algorithm.

    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate 'a priori' priorities). First it applies
    :class:`.DASTB`, and then :class:`.PI`.
    """
    def __init__(self, *args, **kwargs):
        """Saves arguments for :class:`.PI`.

        :param kwargs: Arguments for :class:`.PI`
        """
        self.args = args
        self.kwargs = kwargs

    def run(self, students, schools, ruleset):
        """Runs the algorithm.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """
        DASTB().run(students, schools, ruleset)
        PI(stable_improvements=False, *self.args, **self.kwargs).run(students, schools, ruleset)



"""This module implements the algorithm *'Deferred Acceptance 
with single tie breaking'*.
This implementation is based on Erdil and Ergin (2006)."""

import random

from py_school_match.entities.characteristic import Characteristic
from py_school_match.entities.criteria import Criteria
from py_school_match.entities.rule import Rule

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm
from py_school_match.algorithms.da import DA


class DASTB(AbstractMatchingAlgorithm):
    """This class implements the *'Deferred Acceptance with single
    tie breaking'* algorithm.

    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate 'a priori' priorities), and adds a 'lottery'
    ticket to each student, as an additional rule.
    After this, it executes 'Deferred Acceptance' (from :class:`.DA`).
    """

    def __init__(self):
        """Adds a new :class:`.Criteria`"""
        self.lottery = Criteria("lottery", int)

    def run(self, students, schools, ruleset):
        """Runs the *Deferred Acceptance with multiple tie breaking* algorithm.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """
        lottery_rule = Rule(self.lottery)
        ruleset.add_rule(lottery_rule)

        self.break_ties(students)
        DA().run(students, schools, ruleset)

        ruleset.remove_rule(lottery_rule)

    def break_ties(self, students):
        """This breaks priority ties by assigning lottery tickets.

        Each student is given one lottery ticket (this ticket is created
        as a :class:`.Characteristic`).
        """
        n_students = len(students)
        tickets = random.sample(range(n_students), n_students)

        for index, student in enumerate(students):
            student.add_characteristic(Characteristic(self.lottery, tickets[index]))





"""This package implements a series of matching algorithms.

All algorithms should implement :class:`.AbstractMatchingAlgorithm`.
"""

from py_school_match.algorithms.da_stb import DASTB

from py_school_match.algorithms.da_mtb import DAMTB

from py_school_match.algorithms.sic import SIC
from py_school_match.algorithms.msic import MSIC
from py_school_match.algorithms.nsic import NSIC
from py_school_match.algorithms.mnsic import MNSIC

from py_school_match.algorithms.ttc import TTC

from py_school_match.algorithms.da import DA
from py_school_match.algorithms.pi import PI



"""This module defines a rule."""


class Rule:
    """This class defines rule. 
    A rule is used to determine the priority
    of a student.
    """

    def __init__(self, criteria, quota=1, strict_quota=False):
        """Initializes a rule.

        :param criteria: The criteria this rule checks.
        :type criteria: :class:`.Criteria`.
        :param quota: The minimum percentage a school can have of
        students with the selected criteria.
        :type quota: float.
        :param strict_quota: Whether the quota is strict or flexible.
        A flexible quota means that if there are no more students
        with the selected criteria left, the school can accept
        other students.
        :type strict_quota: bool.
        """
        self.criteria = criteria
        self.quota = quota
        self.strict_quota = strict_quota

    def has_quota(self):
        """If the rule has a quota associated."""
        return self.quota < 1

    def has_strict_quota(self):
        """If the rule has a strict quota associated."""
        return self.quota < 1 and self.strict_quota

    def has_flexible_quota(self):
        """If the rule has a flexible quota associated."""
        return not self.has_strict_quota()





"""This package implements a series of entities which are
necessary to define any matching problem."""

from py_school_match.entities.characteristic import Characteristic
from py_school_match.entities.criteria import Criteria
from py_school_match.entities.ruleset import RuleSet
from py_school_match.entities.rule import Rule

from py_school_match.entities.school import School
from py_school_match.entities.slot import Slot

from py_school_match.entities.social_planner import SocialPlanner

from py_school_match.entities.student import Student
from py_school_match.entities.student_queue import StudentQueue



"""This module defines a queue of students."""

from math import ceil

from py_school_match.entities.slot import Slot


class StudentQueue:
    """This class defines a queue of students."""

    def __init__(self, school, preference_mode=False):
        """Initializes the queue.

        :param school: The school associated.
        :type school: :class:`.School`..
        :param preference_mode: If True, ignores all quotas and capacities.
        It is used to model preferences.
        :type preference_mode: bool.
        """
        self.slot_queue = []
        self.school = school
        self.rejected_student = None

        self.preference_mode = preference_mode

        self.define_slots()

    def get_all_students(self):
        """Returns all the students associated."""
        assigned_students = []
        for slot in self.slot_queue:
            assigned_students.extend([student for student in slot.queue])

        return assigned_students

    def define_slots(self):
        """Creates the slots associated."""
        capacity_used = 0
        if self.school.get_ruleset() and not self.preference_mode:
            for rule in self.school.get_ruleset().get_rules_prioritized():
                if rule.has_quota():
                    slot_size = ceil(rule.quota*self.school.capacity)
                    self.slot_queue.append(Slot(rule, slot_size))
                    capacity_used += slot_size

        self.slot_queue.append(Slot(None, self.school.capacity - capacity_used))

    def append(self, student):
        """Appends a student to the queue."""

        unassigned_student = student
        for slot in self.slot_queue:

            if slot.is_general_slot():
                self.assign_to_slot(slot, unassigned_student)
                unassigned_student = slot.pop_rejected() if not self.preference_mode else None

            elif slot.rule.has_flexible_quota() or slot.rule.criteria in unassigned_student.get_characteristics():
                self.assign_to_slot(slot, unassigned_student)
                unassigned_student = slot.pop_rejected() if not self.preference_mode else None

            self.rejected_student = unassigned_student

            if unassigned_student is None:
                break

    def assign_to_slot(self, slot, unassigned_student):  # ToDo: Possibly move this method to Slot
        """Assigns a student to a slot."""
        slot.append(unassigned_student)
        slot.sort_slot(self.get_priority)

    def get_priority(self, slot, student):
        """Returns the priority of the student inside a slot."""
        priorities = ()
        priority_rules = []

        if self.preference_mode:
            priority_rules = self.school.get_ruleset().get_rules_prioritized()
        else:
            for rule in self.school.get_ruleset().get_rules_prioritized():
                if not rule.has_quota() or (rule == slot.rule and slot.is_quoted_slot() and slot.rule.has_flexible_quota()):
                    priority_rules.append(rule)

        for rule in priority_rules:
            characteristic_group = student.get_characteristic(rule.criteria)  # ToDo: Find a better name for characteristic_group
            if characteristic_group:                                          # It is actually a group of characteristics.
                for characteristic in characteristic_group:
                    if characteristic.is_recognized_in(self.school):
                        rule_priority = self.evaluate_characteristic(characteristic)

                        priorities += (rule_priority,)
                        break
            else:
                priorities += (False,)

        return priorities

    def evaluate_characteristic(self, characteristic):
        """Evaluates a characteristic according to it's criteria associated."""
        if characteristic.criteria.criteria_type == list:
            rule_priority = self.school in characteristic.value
        elif characteristic.criteria.criteria_type == bool:
            rule_priority = characteristic.value
        elif characteristic.criteria.criteria_type == int:
            rule_priority = characteristic.value
        else:
            rule_priority = False

        return rule_priority

    def gen_indifference_groups(self):
        """Returns groups of students with the same priority."""
        groups = []
        for slot in self.slot_queue:
            groups.extend(slot.gen_indifference_groups(self.get_priority))
        return groups

    def get_rejected_student(self):
        """Returns a rejected student (if any)."""
        return self.rejected_student



import os.path
import uuid
import time


def gen_filepath(path_to_file, filename="", extension=None, unique_file=False):
    """Generates a filepath."""
    if extension is None:
        raise Exception("Invalid file extension")

    if path_to_file:
        os.makedirs(path_to_file, exist_ok=True)

    if unique_file or not filename:
        file = "{}_{}.{}".format(time.strftime("%Y%m%d-%H%M%S"), str(uuid.uuid4().hex), extension)
    else:
        file = "{}.{}".format(filename, extension)

    filepath = os.path.join(path_to_file, file)
    return filepath


def get_number_list_from_file(file):
    """Creates a list of numbers from a file (numbers are separated by \n)."""
    with open(file) as f:
        lines = [float(line.rstrip('\n')) for line in f]

    return lines



"""This module implements the algorithm *'Deferred Acceptance with
multiple tie breaking'*, and then searches for stable cycles."""

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm
from py_school_match.algorithms.da_mtb import DAMTB
from py_school_match.algorithms.pi import PI


class MSIC(AbstractMatchingAlgorithm):
    """This class implements the *'Deferred Acceptance with multiple
    tie breaking'* algorithm, and then searches for stable cycles.

    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate 'a priori' priorities). First it applies
    :class:`.DAMTB`, and then :class:`.PI`.
    """
    def __init__(self, *args, **kwargs):
        """Saves arguments for :class:`.PI`.

        :param kwargs: Arguments for :class:`.PI`
        """
        self.args = args
        self.kwargs = kwargs

    def run(self, students, schools, ruleset):
        """Runs the algorithm.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """
        DAMTB().run(students, schools, ruleset)
        PI(stable_improvements=True, *self.args, **self.kwargs).run(students, schools, ruleset)



"""This module defines a social planner."""
from py_school_match.entities.ruleset import RuleSet

class SocialPlanner:
    """This class defines a social planner. 
    A social planner is used to run a algorithm
    using a group of students, a group of schools
    and a set of rules.
    """

    def __init__(self, students, schools, ruleset=None):
        """Initializes a slot.

        :param students: A list of students.
        :type students: list.
        :param schools: A list of school
        :type schools: list.
        :param ruleset: The set of rules. By default, it uses the
        same ruleset in every school.
        :type ruleset: :class:`.Ruleset`.
        """
        self.students = students
        self.schools = schools
        self.ruleset = ruleset if ruleset else RuleSet()

        self.assign_same_ruleset()  # By default it assigns the same ruleset to every school.

    def assign_same_ruleset(self):
        """Assigns the same ruleset to every school."""
        for school in self.schools:
            school.set_ruleset_n_reset(self.ruleset)  # ToDo: The assignation should update itself!

    def run_matching(self, algorithm):
        """Run the algorithm."""
        algorithm.run(self.students, self.schools, self.ruleset)

    @staticmethod
    def change_ruleset(school, ruleset):
        """Changes the ruleset used in a school."""
        school.ruleset = ruleset
        school.reset_assignation()



"""This package implements some util methods.
"""

from py_school_match.util.display_info import get_positions_stat, display_assignation, display_positions_from_planner, display_positions
from py_school_match.util.display_info import display_assignation_from_planner, save_positions, save_data_from_planner

from py_school_match.util.file_util import gen_filepath, get_number_list_from_file

from py_school_match.util.stat import get_variance

from py_school_match.util.simulate_preferences import simulate_preferences
from py_school_match.util.simulate_characteristics import simulate_characteristics


import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='py-school-match',
    version='0.2.0',
    packages=find_packages(exclude=('tests', 'tests.*')),
    url='https://github.com/igarizio/py-school-match',
    license='GLP v3',
    author='Iacopo Garizio',
    author_email='info@iacopogarizio.com',
    description='Py-school-match, a Python package that implements matching algorithms for the student-to-school assignation problem.',
    long_description=read('README.rst')
)



from py_school_match.algorithms import DAMTB, DASTB, SIC, MSIC, NSIC, TTC, MNSIC
from py_school_match.entities import Criteria, Rule, RuleSet, School, SocialPlanner, Student,Characteristic
from py_school_match import util


from random import normalvariate, uniform, seed

from py_school_match.util.stat import get_variance


def simulate_preferences(students, schools, alpha, n_pref_probabilities, schools_factors=None):
    gen_utilities = {}
    for index, school in enumerate(schools):
        gen_utilities[school] = normalvariate(0, 1) if not schools_factors else schools_factors[index]

    for student in students:
        std_utilities = {}
        variance = 1 if not schools_factors else get_variance(schools_factors)
        for school in schools:
            std_utilities[school] = normalvariate(0, variance**0.5) * (1 - alpha) + gen_utilities[school] * alpha

        pref = sorted(std_utilities.keys(), key=lambda s: std_utilities[s], reverse=True)

        n_pref = weighted_choice([(v+1, p) for v, p in enumerate(n_pref_probabilities)])
        student.preferences = pref[:n_pref]


def weighted_choice(choices):
    total = sum(w for c, w in choices)
    r = uniform(0, total)
    upto = 0
    for c, w in choices:
        if upto + w >= r:
            return c
        upto += w
    assert False, "Should not get here"



"""This module defines a student."""

from collections import defaultdict
from itertools import count


class Student:
    """This class defines a student."""

    __id_counter = count(0)
    """This is used to generate an incremental id."""

    def __init__(self, id_=""):
        """Initializes a student.

        :param id_: Any unique identifier.
        :type id_: Any.
        """
        self._id = id_ if id_ else next(Student.__id_counter)
        self.preferences = []

        self.__characteristics = defaultdict(list)

        self.assigned_school = None

    @property
    def id(self):
        return self._id

    @property
    def characteristics(self):
        """Returns a list of characteristics."""
        return self.__characteristics

    def add_characteristic(self, characteristic):
        """Adds a characteristic to the student."""
        self.__characteristics[characteristic.criteria].append(characteristic)

    def get_characteristic(self, criteria):
        """Returns a characteristic associated with a criteria."""
        if criteria in self.__characteristics:
            return self.__characteristics[criteria]
        else:
            return []

    @staticmethod
    def reset_ids():
        """Resets the incremental id."""
        Student.__id_counter = count(0)



def get_variance(elements):
    """Calculates the variance of a list of numbers."""
    m = sum(elements) / len(elements)
    var = sum([(xi - m) ** 2 for xi in elements]) / len(elements)
    return var



"""This module searches for cycles of *'Pareto improvements'*.
It stops when there are no cycles left.

.. DANGER::
   Be careful when running this algorithm. If the number of possible
   improvements is large, this my consume all RAM available.
   If this happens, set:

    .. code-block:: python
       
       USE_STACK_LIMIT = True
       STACK_SAFE_LIMIT = A_CONSTANT
    
    Where :code:`A_CONSTANT` should be a number between 0 and 1.
    
    .. NOTE::
       This may generate a non-optimal solution.
"""

import random
from collections import defaultdict

from graph_tool.all import Graph, graph_draw, all_circuits, sfdp_layout

from py_school_match.algorithms.abstract_matching_algorithm import AbstractMatchingAlgorithm
from py_school_match.entities.student_queue import StudentQueue
from py_school_match.util.file_util import gen_filepath

USE_STACK_LIMIT = True
STACK_SAFE_LIMIT = 0.3


class PI(AbstractMatchingAlgorithm):
    """This class searches for cycles of *'Pareto Improvement'*.

    This takes a list of students, a list of schools and a ruleset
    (which is used to calculate priorities).
    This works by generating a directed graph (with students as edges, 
    and schools as vertices), and then searches for possible cycles;
    if it finds one, it is applied; if not, the algorithm ends.
    """

    EDGE_WIDTH_SIZE_FACTOR = 700
    """Size factor (in the image) of each edge that is not part of the main cycle."""
    EDGE_WIDTH_CYCLE_SIZE = 10
    """Size factor (in the image) of each edge that takes part of the main cycle."""

    def __init__(self, stable_improvements=True, generate_images=False, images_folder="PI_images", use_longest_cycle=True):
        """Runs the algorithm.

        :param stable_improvements: If the algorithm will search for stable improvements or not.
        :type stable_improvements: bool
        :param generate_images: If the process generates images or not.
        :type generate_images: bool
        :param images_folder: Where images are saved.
        :type images_folder: str
        :param use_longest_cycle: If the algorithm applies the longest cycle available, or the first one encountered.
        :type use_longest_cycle: bool
        """
        self.stable_improvements = stable_improvements
        self.generate_images = generate_images
        self.images_folder = images_folder
        self.use_longest_cycle = use_longest_cycle

        self.__graph = None
        self.__vertices_by_id = None
        self.__students_by_id = None
        self.__schools_by_id = None

        self.__school_id = None
        self.__student_id = None

    def reset_variables(self):
        """Resets all variables."""
        self.__graph = Graph()
        self.__vertices_by_id = {}
        self.__students_by_id = {}
        self.__schools_by_id = {}

        self.__school_id = self.__graph.new_vertex_property("int")
        self.__graph.vertex_properties["school_id"] = self.__school_id

        self.__student_id = self.__graph.new_edge_property("int")
        self.__graph.edge_properties["student_id"] = self.__student_id

    def run(self, students, schools, ruleset):
        """Runs the algorithm.
        First it creates the graph, then it lists all the cycles available,
        after that it selects one cycle, and applies it. Finally, it starts
        the process again.

        :param students: List of students.
        :type students: list
        :param schools: List of school.
        :type schools: list
        :param ruleset: Set of rules used.
        :type ruleset: Ruleset
        """

        self.reset_variables()

        can_improve = True
        iteration_counter = 1

        while can_improve:
            self.structure_graph(students, schools)

            cycles = [c for c in all_circuits(self.__graph, unique=True) if len(c) > 1]

            if cycles:
                selected_cycle = max(cycles, key=lambda cycle: len(cycle)) if self.use_longest_cycle else cycles[0]
                cycle_edges = []

                for current_v_index in range(len(selected_cycle)):
                    next_v_index = (current_v_index + 1) % len(selected_cycle)

                    from_v = self.__graph.vertex(selected_cycle[current_v_index])
                    target_v = self.__graph.vertex(selected_cycle[next_v_index])
                    edge = self.__graph.edge(from_v, target_v)
                    cycle_edges.append(edge)

                    edge_student_id = self.__student_id[edge]
                    edge_student = self.__students_by_id[edge_student_id]

                    vertex_school_target_id = self.__school_id[target_v]
                    vertex_school_target = self.__schools_by_id[vertex_school_target_id]

                    if vertex_school_target in edge_student.preferences:
                        edge_student.assigned_school = vertex_school_target

                if self.generate_images:
                    self.generate_image(cycle_edges, iteration_n=iteration_counter)

            else:
                can_improve = False

            self.__graph.clear()
            iteration_counter += 1

    def structure_graph(self, students, schools):
        """Creates a graph where students are edges, and schools vertices.
        
        A vertex is created between two edges, if there is at least one student who
        was assigned to one of the edges and prefers the other. Also, if 
        :code:`stable_cycles` is selected, the student needs to be in the group of
        highest priority of the edge which he prefers.
        
        :param students: List of students.
        :type students: list
        :param schools: 
        :type schools: list
        """
        if not self.__students_by_id and not self.__schools_by_id:  # Maybe this could be moved to Student and School.
            for student in students:
                self.__students_by_id[student.id] = student
            for school in schools:
                self.__schools_by_id[school.id] = school

        for school in schools:
            setattr(school, 'preferences', StudentQueue(school, preference_mode=True))

        for student in students:
            for pref_school in student.preferences:
                if student.assigned_school:
                    try:
                        if student.preferences.index(pref_school) < student.preferences.index(student.assigned_school):
                            pref_school.preferences.append(student)
                    except Exception:
                        print("EXCEPTION", student.id, "---", pref_school.id, student.assigned_school.id, [s.id for s in student.preferences])
                else:
                    pref_school.preferences.append(student)

        for target_school in schools:
            target_school_indiference_groups = target_school.preferences.gen_indifference_groups()
            selected_groups = target_school_indiference_groups[:1] if self.stable_improvements else target_school_indiference_groups

            for group in selected_groups:
                for student in group:
                    stack_safe_limit = True if not USE_STACK_LIMIT else random.random() < STACK_SAFE_LIMIT
                    if student.assigned_school and stack_safe_limit:
                        if student.preferences.index(target_school) < student.preferences.index(student.assigned_school):
                            v_source_school = self.create_vertex(student.assigned_school)
                            v_target_school = self.create_vertex(target_school)
                            self.create_edge(student, v_source_school, v_target_school)

    def create_vertex(self, school):
        """Creates a new vertex in the graph (if it did not existed before)."""
        if school.id in self.__vertices_by_id:
            vertex = self.__vertices_by_id[school.id]
        else:
            vertex = self.__graph.add_vertex()
            self.__vertices_by_id[school.id] = vertex
            self.__school_id[vertex] = school.id
        return vertex

    def create_edge(self, student, v_source_school, target_v):
        """Creates a directed edge between two vertices."""
        e_student = self.__graph.add_edge(v_source_school, target_v)
        self.__student_id[e_student] = student.id

    def generate_image(self, cycle_edges, iteration_n=0):
        """Generates an image of a graph.
        
        :param cycle_edges: Edges which are part of the main cycle (they will be highlighted in red).
        :type cycle_edges: list
        :param iteration_n: Number of iteration of the algorithm (this is added in the filename of the image).
        :type iteration_n: int

        .. DANGER::
        This is an experimental feature.
        """
        edge_width = self.__graph.new_edge_property("int")
        self.__graph.edge_properties["edge_width"] = edge_width

        edge_color = self.__graph.new_edge_property("vector<float>")
        self.__graph.edge_properties["edge_color"] = edge_color

        edge_count = defaultdict(int)

        for edge in self.__graph.edges():
            corresponding_edge = next((e for e in edge_count if self.compare_edges(e, edge)), edge)
            edge_count[corresponding_edge] += 1

        for edge in self.__graph.edges():
            if edge in cycle_edges:
                edge_width[edge] = PI.EDGE_WIDTH_CYCLE_SIZE
                edge_color[edge] = [1., 0.2, 0.2, 0.999]
            else:
                edge_width[edge] = 0.9 + PI.EDGE_WIDTH_SIZE_FACTOR * edge_count[edge] / self.__graph.num_edges()
                edge_color[edge] = [0., 0., 0., 0.3]

        pos = sfdp_layout(self.__graph, K=500, C=20, p=10, theta=20, gamma=0.1)
        graph_draw(self.__graph, pos=pos, vertex_text=self.__school_id, vertex_font_size=30,
                   vertex_fill_color=[0.97, 0.97, 0.97, 1], vertex_color=[0.05, 0.05, 0.05, 0.95],
                   edge_pen_width=edge_width, edge_color=edge_color,
                   output_size=(1000, 1000), bg_color=[1., 1., 1., 1], output=self.generate_filename(iteration_n))

    def generate_filename(self, iteration_n): # ToDo: Move this to utils
        """Returns a filename (which is used to generate the images)."""
        filename = "Graph (iteration {})".format(iteration_n) if iteration_n > 0 else "Graph"
        output_file = gen_filepath(self.images_folder, filename=filename, extension="png")
        return output_file

    def compare_edges(self, e1, e2):
        """Compares two edges."""
        return (self.__school_id[e1.source()] == self.__school_id[e2.source()]) and (self.__school_id[e1.target()] == self.__school_id[e2.target()])

